import copy
from contextlib import contextmanager
from dataclasses import dataclass
import typing
import warnings

import numpy as np
import pandas as pd
import torch as t
import torch.nn as nn
import xarray as xr
from einops import rearrange
from bidict import bidict

# TODO: simplify dependencies for visualizing computational graph
# import panel as pn
# pn.extension('plotly', comms='ipywidgets')
# import hvplot.networkx as hvnx
# import networkx as nx
# from networkx.algorithms import bipartite
# import holoviews as hv

@dataclass
class ModuleHookData():
    inputs:      tuple
    output:      t.Tensor
    custom_data: typing.Any

@dataclass
class PatchDef():
    '''Both tensors should be the size of the activation value to be patched, 
    with the first batch dimension optionally length-1 to broadcast in this dim.'''
    mask:  t.Tensor  # True where patched values should be applied
    value: t.Tensor  # Patch values

class ModuleHook():
    network: nn.Module
    values_by_label: bidict[str, t.Tensor]
    modules_by_label: dict[str, nn.Module]
    module_labels_by_id: dict[int, str]
    module_data_by_label: dict[str, typing.Any]
    meta_by_label: dict[str, typing.Any]
    value_to_module_edges: set[tuple[str, str]]
    module_to_value_edges: set[tuple[str, str]]

    def __init__(self, network, module_labels=None):
        '''Copy the provided network and add forward hooks to extract values from all layers
        during subsequent forward calls.  Also creates data structures to hold module input/output
        values and relationships to allow building of the module graph.'''
        # Copy network
        self.network = copy.deepcopy(network)
        # Init data structures
        self.modules_by_label = {}
        self.module_labels_by_id = {}
        self.module_data_by_label = {}
        self.meta_by_label = {}
        self.values_by_label = bidict()
        self.value_to_module_edges = set()
        self.module_to_value_edges = set()
        self.module_data_by_label = {}
        # Register all the modules with labels
        if module_labels is not None:
            for label, mod in zip(module_labels, self.network.modules()):
                self.register_module(mod, label)
        else:
            # Create module labels based on internal module variable names if not provided
            for label, mod in self.network.named_modules():
                self.register_module(mod, label)
        # Attach the forward hook function to all modules in the provided network
        self.network.apply(lambda m: self.register_hook(m))
        # Init some variables
        self.hook_should_get_custom_data = False
        self.batch_index = None
        self.last_batch_index = None
        self.patches = {}

    def register_hook(self, module):
        # Register forward hook on provided module
        module.register_forward_hook(self._module_hook)

    @contextmanager
    def set_hook_should_get_custom_data(self):
        self.hook_should_get_custom_data = True
        yield None
        self.hook_should_get_custom_data = False

    @contextmanager
    def use_patches(self, patches):
        self.patches = patches
        yield None
        self.patches = {}

    @contextmanager
    def use_batch_index(self, batch_index):
        self.batch_index = batch_index
        yield None
        self.batch_index = None

    def tensor_to_result_object(self, tensor, label):
        array = tensor.detach().numpy()
        if self.last_batch_index is None:
            return array
        dims = [self.last_batch_index.name] + ['{}_d{}'.format(label, ii) for ii in range(1, len(array.shape))]
        coords = {dims[0]: self.last_batch_index}
        for dim, ln in list(zip(dims, array.shape))[1:]:
            coords[dim] = np.arange(ln)
        return xr.DataArray(array, dims=dims, coords=coords)

    def register_module(self, module, label):
        '''If a module isn't registered already, add it to the various data structures.'''
        _id = id(module)
        if _id not in self.module_labels_by_id:
            self.module_labels_by_id[_id] = label
            self.modules_by_label[label] = module
            self.meta_by_label[label] = dict(desc=' '.join(str(module).split()))
        return self.module_labels_by_id[_id]

    def set_module_data(self, label, module_data):
        self.module_data_by_label[label] = module_data

    def set_value(self, label: str, value: t.Tensor):
        # self.values_by_label is a bidict, so keys (labels) and values (value
        # tensors) are both unique sets.
        # First, check if this value object is already in the bidict; if so,
        # return it's label (value labels shouldn't ever change)
        if value in self.values_by_label.inverse:
            return self.values_by_label.inverse[value]
        # Otherwise, set the provided label to point to this value
        self.values_by_label[label] = value
        try:
            desc = str(value.shape) # Value is a single tensor
        except AttributeError:
            # Value might be a tuple of tensors?
            try:
                desc = '({})'.format(', '.join([str(vv.shape) for vv in value]))
            except AttributeError:
                desc = ''
        self.meta_by_label[label] = dict(desc=desc)
        return label

    def _module_hook(self, module, inps, outp):
        # Store the batch index that should apply to current values
        self.last_batch_index = self.batch_index
        # Get the label for this module
        module_label = self.module_labels_by_id[id(module)]
        # Store the input values
        inp_labels = []
        for ii, inp in enumerate(inps):
            inp_labels.append(self.set_value("{}_in{}".format(module_label, ii), inp))
        # Store the output value
        outp_label = self.set_value("{}_out".format(module_label), outp)
        # Store edges to this module
        for inp_label in inp_labels:
            self.value_to_module_edges.add((inp_label, module_label))
        # Store output edge from this module, but only if the output object is 
        # different to all input objects (i.e. avoid cycles if we have identity modules)
        if not id(outp) in [id(inp) for inp in inps]:
            self.module_to_value_edges.add((module_label, outp_label))        

        # Populate custom data, if enabled (this can be large)
        if self.hook_should_get_custom_data:
            # Create module data object
            # Convert input and output objects to appropriate type depending on
            # whether an index is set
            module_data = ModuleHookData(inps, outp, None)
            # Add custom data for this module based on module type
            if isinstance(module, nn.Linear):
                module_data.custom_data = dict(weight=module.state_dict()['weight'],
                    bias=module.state_dict()['bias'])
                if inps[0].shape[0] == 1:
                    # Include pre-sum data if we're only looking at a single input point,
                    # otherwise could use too much memory?
                    presum = inps[0] * module.state_dict()['weight']
                    module_data.custom_data['presum'] = presum
            if isinstance(module, nn.Conv2d):
                weight = module.state_dict()['weight']
                module_data.custom_data = dict(weight=weight,
                    bias=module.state_dict()['bias'])
                if inps[0].shape[0] == 1:
                    # Include pre-sum data if we're only looking at a single input point,
                    # otherwise could use too much memory?
                    # Pre-calculate the by-channel convolutions before the final sums
                    # Weight tensor dims are (out_channel, in_channel, kh, kw)
                    # We can calculate the intermediate presum values by running a single
                    # conv2d call for each input, but across all outputs
                    # TODO: handle groups argument
                    # TODO: consider making all custom data (e.g. presums, etc. xarrays to
                    # make plotting easier)
                    assert module.groups == 1, 'Function only supports groups=1 for now'
                    presum_list = []
                    for inch in range(weight.shape[1]):
                        presum_this_input = nn.functional.conv2d(
                            inps[0][:,inch:(inch+1),:,:], weight=weight[:,inch:(inch+1),:,:], 
                            stride=module.stride, padding=module.padding, 
                            dilation=module.dilation)
                        presum_list.append(presum_this_input)
                    presum_tensor = rearrange(presum_list, "inch b outch h w -> inch outch b h w")
                    module_data.custom_data['presum'] = presum_tensor
            # Store the module data
            self.set_module_data(module_label, module_data)
            
        # Handle patching, if any
        if outp_label in self.patches:
            patch = self.patches[outp_label]
            if isinstance(patch, PatchDef):
                pmask = patch.mask
                pvalue = patch.value
                outp = outp*(~pmask) + pvalue*pmask
            else:
                # Patch must be a custom function
                outp = patch(outp)
            #for slice_tuple, patched_value in self.patches[outp_label]:
            #    outp[slice_tuple] = t.Tensor(patched_value)
            # Update the stored value since it's changed
            self.set_value("{}_out".format(module_label), outp)
            return outp

    def probe_with_input(self, inp, patches={}, func=None, **kwargs):
        warnings.warn('''Function probe_with_input is deprecated due to
        a confusing name and has been replaced with run_with_input.''')
        return self.run_with_input(inp, patches, func, **kwargs)
    
    def run_with_input(self, inp, patches={}, func=None, **kwargs):
        '''Makes a forward pass over the network, creating all intermediate module
        input/output values.  By default, customizes the forward call to use predict() 
        if module type supports it, so that we get all the observation pre-processing,
        etc, otherwise uses a normal forward call.  The function to be called to
        execute the forward pass can also be overriden using the func argument,
        which will be called with the hooked network, input as a tensor and **kwargs.
        Additional keyword arguments will be passed to the forward call in all cases.'''
        # Make sure the input is an ndarray at this stage, but if DataArray is passed,
        # all hook results will also be DataArrays with same batch dimension (assumed
        # to be first dimension implicitly here), else numpy ndarrays.
        batch_index = None
        if isinstance(inp, xr.DataArray):
            batch_index = inp.indexes[inp.dims[0]]
            inp = inp.to_numpy()
        # Make forward pass using patches, batch index (if any), and storing intermediate data
        with self.set_hook_should_get_custom_data(), self.use_patches(patches), \
                self.use_batch_index(batch_index), t.no_grad():
            # If a function is provided, call it with the input as a tensor
            if func is not None:
                outp = func(self.network, t.from_numpy(inp), **kwargs)
            else:
                # Default to using predict() if present, otherwise normal forward call
                try:
                    # predict() wants an ndarray
                    outp = self.network.predict(inp, **kwargs)
                except AttributeError:
                    # Normal module forward call wants a tensor
                    outp = self.network(t.from_numpy(inp), **kwargs)
        return outp        
            

    def get_module_and_data_by_label(self, label):
        module = self.modules_by_label[label]
        data = self.module_data_by_label[label]
        return module, data

    def get_value_by_label(self, label, convert=True):
        value = self.values_by_label[label]
        if convert:
            return self.tensor_to_result_object(value, label)
        return value

    def get_preceders(self, label, include_parent_modules=False):
        '''Step backward through computational graph and return preceding
        module/value label.'''
        # TODO: make this more efficient, prob shouldn't need to loop through edges?
        preceders = []
        for n1, n2 in (list(self.module_to_value_edges) + list(self.value_to_module_edges)):
            if n2 == label:
                preceders.append(n1)
        if not include_parent_modules:
            preceders = [prec for prec in preceders if not (prec in self.modules_by_label and
                len(list(self.modules_by_label[prec].children()))!=0)]
        return preceders

    # TODO: simplify dependencies
    # def get_graph_after_probe(self, include_parent_modules=True, show_unconnected=False):
    #     # Nodes and edges using labels
    #     def make_node_tuples(labels):
    #         return [(label, self.meta_by_label[label]) for label in labels]
    #     value_nodes = make_node_tuples(self.values_by_label.keys())
    #     modules_with_edges = set([mod for val, mod in self.value_to_module_edges] +
    #         [mod for mod, val in self.module_to_value_edges])
    #     modules_to_use = set([label for label, mod in self.modules_by_label.items() 
    #         if ((include_parent_modules or len(list(mod.children()))==0) and
    #             (show_unconnected       or label in modules_with_edges)     )])
    #     module_nodes = make_node_tuples(modules_to_use)
    #     edges = [(mod, val) for mod, val in self.module_to_value_edges if mod in modules_to_use] + \
    #         [(val, mod) for val, mod in self.value_to_module_edges if mod in modules_to_use]
    #     # Build the graph
    #     B = nx.DiGraph()
    #     B.add_nodes_from(module_nodes, type='module')
    #     B.add_nodes_from(value_nodes, type='value')
    #     B.add_edges_from(edges)
    #     # Lay out the graph
    #     pos = nx.layout.kamada_kawai_layout(B)
    #     # Calculate node colors 
    #     node_top_level_by_label = {lab: ii for ii, lab in enumerate(list(nx.topological_sort(B)))}
    #     node_top_levels = [node_top_level_by_label[lab] for lab in B.nodes]
    #     # Create the plot
    #     viz = hvnx.draw(B, pos, node_color=node_top_levels, cmap='Blues', 
    #         node_size=200-(hv.dim('type')=='value')*100, arrowhead_length=0.01 ,width=800, height=800)
    #     # Add the responsive tab object
    #     stream_selection = hv.streams.Selection1D(source=viz.nodes)
    #     @pn.depends(stream_selection.param.index)
    #     def selection(index):
    #         if len(index) == 0:
    #             ss = 'No selection'
    #         else:
    #             ii = index[0]
    #             label = list(B.nodes)[ii]
    #             obj = self.objects_by_label[label]
    #             _id = id(obj)
    #             ss = '\n'.join([
    #                 'Node index: {}'.format(ii),
    #                 'Label: {}'.format(label),
    #                 'Type: {}'.format('module' if isinstance(obj, nn.Module) else 'value'),
    #                 'ID: {}'.format(_id)])
    #         return pn.pane.Str(ss, width=200)
    #     #return pn.Column(viz.nodes.opts(alpha=0) * viz, selection)
    #     return pn.Pane(viz)


# %%
import os
import typing
from dataclasses import dataclass
import warnings
import datetime
import json
#import lz4.frame
#import lzma
#import bz2
#import gzip
import blosc
import pickle

import numpy as np
import pandas as pd
import torch as t
import torch.nn as nn
import xarray as xr
import gym
from tqdm.auto import tqdm
from moviepy.video.io.ImageSequenceClip import ImageSequenceClip
from einops import rearrange

os.environ["SDL_VIDEODRIVER"] = "dummy"

@dataclass
class RlStepSeq():
    renders:      xr.DataArray    # The RGB or grayscale outputs of the environment rendering at each sequence point
    obs:          xr.DataArray  
    actions:      xr.DataArray  
    model_states: typing.Optional[xr.DataArray]
    rewards:      xr.DataArray  
    dones:        xr.DataArray
    custom:       dict[str, xr.DataArray]

def run_rollout(
        model_or_func, 
        env: gym.Env, 
        deterministic: bool=False, 
        seed: int=1, 
        max_episodes: typing.Optional[int]=1,
        max_steps: typing.Optional[int]=None, 
        show_pbar: bool=True,
        custom_data_funcs: dict={}) -> typing.Tuple[RlStepSeq, float, int]:
    '''Run episodes, using provided policy in provided environment.
    Save and return episode data as an RlStepSeq object.'''
    assert max_episodes is not None or max_steps is not None, "Must provide max episodes or steps."
    # Confirm that the environment is vectorized (required)
    assert hasattr(env, 'num_envs'), "Environment must be vectorized"
    # This function only works for a single environment
    assert env.num_envs == 1, "Only single environment is supported."
    # Determine what function to call for getting actions
    if isinstance(model_or_func, nn.Module):
        if hasattr(model_or_func, 'predict'):
            predict_func = model_or_func.predict
        else:
            predict_func = model_or_func.forward
    else:
        predict_func = model_or_func
    # Init all the seeds!
    t.manual_seed(seed)
    t.cuda.manual_seed_all(seed)
    t.backends.cudnn.benchmark = False
    t.backends.cudnn.deterministic = True
    env.action_space.seed(seed)
    env.observation_space.seed(seed)
    try:
        env.seed(seed)
    except (TypeError, AttributeError):
        warnings.warn('Seed setting failed.')
    # Reset the environment to get the initial observation
    obs = env.reset()
    # Init lists to store all the data since we don't know the number of steps yet
    renders_all = []
    obs_all = []
    actions_all = []
    model_states_all = []
    rewards_all = []
    dones_all = []
    custom_data_all = {nm: [] for nm in custom_data_funcs.keys()}
    step_cnt = 0
    is_complete = False
    episode_returns = []
    with tqdm(total=max_steps, disable=not show_pbar) as pbar:
        while not is_complete:
            # Convert the observation from dict form if needed
            if isinstance(obs, dict):
                obs = obs['rgb']
            # Get the latest frame
            render = env.render(mode='rgb_array')
            # Get action and model states
            with t.no_grad():
                action, model_states, *extra = predict_func(obs, deterministic=deterministic)
            # Vectorize model states if it's None for consistent interface
            model_states = [None] if model_states is None else model_states
            # Call any provided custom data functions (these are called before the env step,
            # so we have env, current render/obs, agent action/state, but no reward/done/obs_next)
            for nm, func in custom_data_funcs.items():
                custom_data_all[nm].append(
                    func(env=env, render=render, obs=obs[0], action=action[0], 
                        model_states=model_states[0], predict_extra=extra))
            # Step environment
            obs_next, rewards, dones, infos = env.step(action)
            # Store the obs, action, model_state, reward, done
            renders_all.append(render)
            obs_all.append(obs[0])
            actions_all.append(action[0])
            model_states_all.append(model_states[0])
            rewards_all.append(rewards[0])
            dones_all.append(dones[0])
            step_cnt += env.num_envs
            # Detect whether episode has terminated
            # TODO: done sometimes set when losing a life in Atari games? Use something else here?
            if dones[0]:
                episode_returns.append(infos[0].get('episode', {'r':0})['r'])
                # Terminate after max episodes if provided
                if max_episodes is not None and len(episode_returns) >= max_episodes:
                    is_complete = True
            # Terminate after max steps if provided
            if max_steps is not None and step_cnt >= max_steps:
                is_complete = True
            # Move to new observation for next loop
            obs = obs_next
            # Tick the progress bar
            pbar.update(1)
    # Wrap every up and return
    env.close()
    # Make the sequence data object
    # Renders
    # # TEMP: debug, hackily show frame index as a pixel inversion!
    # for ii, render in enumerate(renders_all):
    #     ww = ii % render.shape[1]
    #     render[10,ww,:] = 255 - render[10,ww,:]
    coords = {'step': np.arange(len(renders_all))}
    renders_da = xr.DataArray(rearrange(renders_all, 'step h w rgb -> step h w rgb'), 
        dims=('step', 'h', 'w', 'rgb'), coords=coords)
    # Observations
    obs_da = xr.DataArray(rearrange(obs_all, 'step ... -> step ...'), 
        dims=tuple(['step']+['obd{}'.format(ii) for ii, sz in 
                enumerate(obs_all[0].shape) if sz > 1]), coords=coords)
    # Actions
    actions_da = xr.DataArray(np.array(actions_all), dims=('step'), coords=coords)
    # Model states
    if model_states_all[0] is not None:
        model_states_da = xr.DataArray(rearrange(model_states_all, 'step ms -> step ms'), 
            dims=('step', 'model_state'), coords=coords)
    else:
        model_states_da = None
    # Rewards
    rewards_da = xr.DataArray(np.array(rewards_all), dims=('step'), coords=coords)
    # Dones
    dones_da = xr.DataArray(np.array(dones_all), dims=('step'), coords=coords)
    # Custom data
    custom_data_das = {}
    for nm, custom_data in custom_data_all.items():
        if isinstance(custom_data[0], np.ndarray):
            custom_data_das[nm] = xr.DataArray(rearrange(custom_data, 'step ... -> step ...'), 
                dims=tuple(['step']+['{}_d{}'.format(nm, ii) for ii, sz in 
                        enumerate(custom_data[0].shape)]), coords=coords)
        else:
            custom_data_das[nm] = xr.DataArray(np.array(custom_data), 
                dims=('step'), coords=coords)
    # Final combined data object
    seq_data = RlStepSeq(renders_da, obs_da, actions_da, model_states_da, rewards_da, dones_da, custom_data_das)
    return seq_data, episode_returns, step_cnt

def split_seq(seq: RlStepSeq, start_inds):
    '''Split sequence object into subsequences, with start_inds defining the start indices,
    of each sequence.  Sequences will span from start_inds[i] to start_inds[i+1]-1 inclusive.
    There will be len(start_inds)-1 subsequences returned.'''
    new_seqs = []
    for start_ind, end_ind in zip(start_inds[:-1], start_inds[1:]):
        args = [aa.sel(step=slice(start_ind, end_ind)) if aa is not None else aa for aa in 
            [seq.renders, seq.obs, seq.actions, seq.model_states, seq.rewards, seq.dones]]
        new_seq = RlStepSeq(*args)
        new_seqs.append(new_seq)
    return new_seqs

def make_video_from_renders(renders: xr.DataArray, fps=30.):
    vid_fn = "temp_seq_video.mp4"
    clip = ImageSequenceClip([aa.to_numpy() for aa in renders], 
        fps=fps)
    clip.write_videofile(vid_fn, logger=None)
    return vid_fn, fps

class TempVideoFileFromSeq:
    def __init__(self, seq: RlStepSeq, fps=30.):
        self.seq = seq
        self.fps = fps
    def __enter__(self):
        # Create the temp file from the renders, return the filename
        self.vid_fn, _ = make_video_from_renders(seq.renders, fps=self.fps)
    def __exit__(self, *args, **kwargs):
        # Delete the temp file if it exists
        if os.path.isfile(self.vid_fn):
            os.remove(self.vid_fn)

# Function to generate dataset of episodes
def make_dataset(predict_func, desc, path, num_episodes, 
        env_setup_func,
        seq_mod_func=lambda seq: seq, 
        run_rollout_kwargs={}):
    # Create a timestamped output folder
    utc = datetime.datetime.utcnow()
    dr = utc.strftime(os.path.join(path, '%Y%m%dT%H%M%S'))
    os.mkdir(dr)
    # Create a dataset metadata file
    metadata = dict(
        desc = desc,
        timestamp = utc.strftime('%Y%m%dT%H%M%S'),
        num_episodes = num_episodes,
        run_rollout_kwargs = run_rollout_kwargs
    )
    with open(os.path.join(dr, 'metadata.json'), 'w') as fl:
        json.dump(metadata, fl, indent=2, default=lambda a: str(a))
    # Loop through episodes
    ep_cnt = 0
    for ep_cnt in tqdm(range(num_episodes)):
        # Create env
        venv, episode_metadata = env_setup_func()
        # Run rollout
        #run_rollout_kwargs['show_pbar'] = False
        seq, episode_returns, step_cnt = run_rollout(predict_func, venv, **run_rollout_kwargs)
        # Prepare episode data object
        episode_data = dict(
            episode_metadata=episode_metadata,
            seq=seq_mod_func(seq),
            step_cnt=step_cnt,
            ep_cnt = ep_cnt,
        )
        # Save episode data
        #with open(os.path.join(dr, f'{ep_cnt}.pk'), "wb") as fl:
        #with lzma.open(os.path.join(dr, f'{ep_cnt}.xa'), "wb") as fl:
        #with lz4.frame.open(os.path.join(dr, f'{ep_cnt}.lz4'), "wb") as fl:
        #with bz2.BZ2File(os.path.join(dr, f'{ep_cnt}.bz2'), "wb") as fl:
        #with gzip.open(os.path.join(dr, f'{ep_cnt}.gz'), "wb") as fl:
        #    pickle.dump(episode_data, fl, protocol=pickle.HIGHEST_PROTOCOL)

        pickled_data = pickle.dumps(episode_data, protocol=pickle.HIGHEST_PROTOCOL)
        compressed_pickle = blosc.compress(pickled_data)
        with open(os.path.join(dr, f'{ep_cnt}.dat'), "wb") as fl:
            fl.write(compressed_pickle)

def load_saved_rollout(fn):
    with open(fn, "rb") as f:
        compressed_pickle = f.read()
    depressed_pickle = blosc.decompress(compressed_pickle)
    return pickle.loads(depressed_pickle)

# %%
# Basic tests

if __name__ == "__main__":
    # Dummy predict function
    def predict(obs, deterministic=False):
        return np.array([0]), None

    # Test with a normal gym env
    # env = gym.make('CartPole-v1')
    # run_rollout(predict, env, max_steps=50)

    # Test an atari env
    # from stable_baselines3.common.vec_env import VecFrameStack
    # from stable_baselines3.common.env_util import make_atari_env
    # env = make_atari_env("PongNoFrameskip-v4", n_envs=1, seed=0)
    # env = VecFrameStack(env, n_stack=4)
    # seq, episode_returns, step_cnt = run_rollout(predict, env, max_steps=50)

    # Test with a procgen gym3 env
    import procgen
    env = procgen.ProcgenEnv(num_envs=1, env_name='maze', num_levels=1, start_level=0)
    seq, episode_returns, step_cnt = run_rollout(predict, env, max_steps=50)


from dataclasses import dataclass

import numpy as np
from einops import rearrange
import xarray as xr
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import f_classif, mutual_info_classif
from sklearn import metrics
from tqdm.auto import tqdm

def f_classif_fixed(X, y, **kwargs):
    '''Handle columns with zero variance, hackily'''
    X_fixed = X
    X_fixed[0,:] += 1e-6
    return f_classif(X_fixed, y, **kwargs)

def get_sort_inds_and_ranks(x):
    sort_inds = x.argsort()
    ranks = np.empty_like(sort_inds)
    ranks[sort_inds] = np.arange(len(x))
    return sort_inds, ranks

def run_probe(hook, target, 
        value_labels = None,
        index_fracs = np.linspace(0, 1.0, 11, endpoint=True),
        index_nums = None,
        rank_method = 'f_test',
        test_size = 0.2,
        random_state = 42,
        do_scale_activations = True,
        target_labels = None,
        logistic_regression_kwargs = {}):
    '''Run a probe to train linear classifiers (possibly multi-class) on
    datasets of activations from a provided list of value labels, given a set
    of target classfier outputs.  These classifiers can optionally be sparse,
    in which case the activations will be ranked using the provided rank_method.
    Either a list of fractions or nums can be provided to determine the top-K
    activations to pull from the activations after they have been ranked.
    The same training subset is used to rank activations and to train 
    all the classifers for a given value label.

    Activation values can be optionally scaled using a StandardScaler before
    training.  Default is yes.

    Target can by an xarray or numpy array

    Hook values must have been created such that their first dimension is 
    the dimension over the dataset batch.
    
    Returns an array of trained classifier objects, with dims 
    (value_label, num_acts)'''
    # TODO: check argument shapes, etc.

    # Make sure target is a numpy array
    if isinstance(target, xr.DataArray):
        target = target.values

    # Create target class labels if not provided
    y = target
    y_unique = np.unique(y)
    num_target_labels = len(y_unique)
    is_multiclass = num_target_labels > 2
    if target_labels is None:
        target_labels = y_unique
    
    # Try all values if none provided
    if value_labels is None:
        value_labels = hook.values_by_label.keys()

    # Sizes of results structure
    num_values = len(value_labels)
    num_index_nums = len(index_nums) if index_nums is not None \
        else len(index_fracs)
    results_shape = (num_values, num_index_nums)
    results_dims = ['value_label', 'value_num_step']

    # Structure to hold the results
    probe_results = xr.Dataset(dict(
        classifier = xr.DataArray(
            data=np.zeros(results_shape, dtype=LogisticRegression),
            dims=results_dims),
        score = xr.DataArray(data=np.zeros(results_shape), dims=results_dims),
        conf_matrix = xr.DataArray(
            data=np.zeros(results_shape+(num_target_labels, num_target_labels), 
                dtype=int),
            dims=results_dims+['test_label', 'pred_label']),
        report = xr.DataArray(data=np.zeros(results_shape, dtype=str), 
            dims=results_dims))).assign_coords(dict(
                value_label = value_labels,
                value_num_step = np.arange(num_index_nums),
                test_label = target_labels,
                pred_label = target_labels,
            ))
            
    # Iterate over value labels
    for label in tqdm(value_labels):
        value = hook.get_value_by_label(label)

        # Create training data, reshaping activation value to 2D
        X = rearrange(value.values, 'b ... -> b (...)')
        D_act = X.shape[1]

        # Scale, if requested
        if do_scale_activations:
            scaler = StandardScaler()
            X_scl = scaler.fit_transform(X)
        else:
            scaler = None
            X_scl = X

        # Split into train and test set
        X_train, X_test, y_train, y_test = train_test_split(X_scl, y, 
            test_size=test_size, random_state=random_state)

        # Get activation ranking using provided method
        if rank_method == 'f_test':
            f_test_train, _ = f_classif_fixed(X_train, y_train)
            sort_inds_train, ranks_train = get_sort_inds_and_ranks(f_test_train)

        # Get the number of activations to use at each step, if not provided
        if index_nums is None:
            index_nums = np.array(index_fracs * D_act).astype(int)
        
        # Iterate over activation nums, training with each set of activations
        for kk, K in enumerate(index_nums):
            # Get the indices of the activations to use
            top_K_inds_train = sort_inds_train[-K:]
            
            # Get the top-K data
            X_top_train = X_train[:,top_K_inds_train]
            X_top_test = X_test[:,top_K_inds_train]

            # Create an appropriate classifier
            if is_multiclass:
                # Multi-class regression
                clf = LogisticRegression(random_state=random_state, 
                    multi_class='ovr', solver='liblinear', 
                    **logistic_regression_kwargs)
            else:
                clf = LogisticRegression(random_state=random_state,
                    **logistic_regression_kwargs)

            # Train!
            clf.fit(X_top_train, y_train)

            # Test
            # (Some useful reference code here: https://www.kaggle.com/code/satishgunjal/multiclass-logistic-regression-using-sklearn/notebook)
            y_pred = clf.predict(X_top_test)
            score = clf.score(X_top_test, y_test)
            # if is_multiclass:
            #     conf_matrix = metrics.multilabel_confusion_matrix(y_test, y_pred, 
            #         labels=target_labels)
            # else:
            #     conf_matrix = metrics.confusion_matrix(y_test, y_pred, labels=target_labels)
            conf_matrix = metrics.confusion_matrix(y_test, y_pred, labels=target_labels)
            report = metrics.classification_report(y_test, y_pred)

            # Collate all the results into an object
            #relevant_indices.loc[dict(value=value_label, count_step=final_count_step)].values[()] = relinds_this
            probe_results['classifier'].loc[
                dict(value_label=label, value_num_step=kk)].values[()] = clf
            probe_results['score'].loc[
                dict(value_label=label, value_num_step=kk)].values[()] = score
            probe_results['conf_matrix'].loc[
                dict(value_label=label, value_num_step=kk)].values[:,:] = conf_matrix
            probe_results['report'].loc[
                dict(value_label=label, value_num_step=kk)].values[()] = report

    # Return the results
    return probe_results